@@ -113,42 +113,53 @@ class FeedForward(nn.Module):
     def forward(self, feats, coors):
         return self.net(feats), 0
 
+class Attention(nn.Module):
+    def __init__(self, dim, heads = 8, dim_head = 64):
+        super().__init__()
+        inner_dim = heads * dim_head
+        self.heads = heads
+        self.scale = dim_head ** -0.5
+
+        self.to_q = nn.Linear(dim, inner_dim, bias = False)
+        self.to_kv = nn.Linear(dim, inner_dim * 2, bias = False)
+        self.to_out = nn.Linear(inner_dim, dim)
+
+    def forward(self, x, context, mask = None):
+        h = self.heads
+
+        q = self.to_q(x)
+        kv = self.to_kv(context).chunk(2, dim = -1)
+
+        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, *kv))
+        dots = einsum('b h i d, b h j d -> b h i j', q, k) * self.scale
+
+        if exists(mask):
+            mask_value = max_neg_value(dots)
+            mask = rearrange(mask, 'b n -> b () () n')
+            dots.masked_fill_(~mask, mask_value)
+
+        attn = dots.softmax(dim = -1)
+        out = einsum('b h i j, b h j d -> b h i d', attn, v)
+
+        out = rearrange(out, 'b h n d -> b n (h d)', h = h)
+        return self.to_out(out)
+
 class GlobalLinearAttention(nn.Module):
     def __init__(
         self,
         *,
         dim,
-        dim_head,
-        heads
+        heads = 8,
+        dim_head = 64
     ):
         super().__init__()
-        inner_dim = dim_head * heads
-        self.heads = heads
-        self.scale = dim_head ** -0.5
-        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias = False)
-        self.to_out = nn.Linear(inner_dim, dim)
+        self.attn1 = Attention(dim, heads, dim_head)
+        self.attn2 = Attention(dim, heads, dim_head)
 
-    def forward(self, feats, coors, mask = None):
-        h = self.heads
-        q, k, v = self.to_qkv(feats).chunk(3, dim = -1)
-        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
-
-        if exists(mask):
-            mask = rearrange(mask, 'b n -> b () n ()')
-            k.masked_fill_(~mask, -torch.finfo(k.dtype).max)
-
-        q = q.softmax(dim = -1)
-        k = k.softmax(dim = -2)
-
-        q = q * self.scale
-
-        if exists(mask):
-            v.masked_fill_(~mask, 0.)
-
-        context = einsum('b h n d, b h n e -> b h d e', k, v)
-        out = einsum('b h d e, b h n d -> b h n e', context, q)
-        out = rearrange(out, 'b h n d -> b n (h d)')
-        return self.to_out(out), 0
+    def forward(self, x, queries, mask = None):
+        induced = self.attn1(queries, x, mask = mask)
+        out     = self.attn2(x, induced)
+        return out, 0
 
 class EquivariantAttention(nn.Module):
     def __init__(
@@ -408,7 +419,8 @@ class EnTransformer(nn.Module):
         valid_neighbor_radius = float('inf'),
         init_eps = 1e-3,
         norm_rel_coors = True,
-        global_linear_attn_every = 0
+        global_linear_attn_every = 0,
+        num_global_tokens = 8
     ):
         super().__init__()
         assert dim_head >= 32, 'your dimension per head should be greater than 32 for rotary embeddings to work well'
@@ -427,12 +439,21 @@ class EnTransformer(nn.Module):
         self.layers = nn.ModuleList([])
 
         global_linear_attn_every = default(global_linear_attn_every, 0)
+        has_global_attn = global_linear_attn_every > 0
+
+        self.global_tokens = None
+        if has_global_attn:
+            self.global_tokens = nn.Parameter(torch.randn(num_global_tokens, dim))
 
         for ind in range(depth):
             add_global = global_linear_attn_every > 0 and (ind % global_linear_attn_every) == 0
 
             self.layers.append(nn.ModuleList([
-                Residual(PreNorm(dim, GlobalLinearAttention(dim = dim, heads = heads, dim_head = dim_head))) if add_global else None,
+                nn.ModuleList([
+                    nn.LayerNorm(dim),
+                    PreNorm(dim, GlobalLinearAttention(dim = dim, heads = heads, dim_head = dim_head)),
+                    Residual(PreNorm(dim, FeedForward(dim = dim))),
+                ])  if add_global else None,
                 Residual(PreNorm(dim, EquivariantAttention(dim = dim, dim_head = dim_head, heads = heads, coors_hidden_dim = coors_hidden_dim, edge_dim = (edge_dim + adj_dim),  neighbors = neighbors, only_sparse_neighbors = only_sparse_neighbors, valid_neighbor_radius = valid_neighbor_radius, init_eps = init_eps, rel_pos_emb = rel_pos_emb, norm_rel_coors = norm_rel_coors))),
                 Residual(PreNorm(dim, FeedForward(dim = dim)))
             ]))
@@ -479,9 +500,20 @@ class EnTransformer(nn.Module):
 
         # main network
 
-        for global_attn, attn, ff in self.layers:
-            if exists(global_attn):
-                feats, coors = global_attn(feats, coors, mask = mask)
+        global_tokens = None
+        if exists(self.global_tokens):
+            global_tokens = repeat(self.global_tokens, 'n d -> b n d', b = b)
+
+        for global_fns, attn, ff in self.layers:
+            if exists(global_fns):
+                global_norm, global_attn, global_ff = global_fns
+
+                feats_out, global_out = global_attn(feats, global_norm(global_tokens), mask = mask)
+
+                feats = feats + feats_out
+                global_tokens = global_tokens + global_out
+
+                feats, coors = global_ff(feats, coors)
 
             feats, coors = attn(feats, coors, edges = edges, mask = mask, adj_mat = adj_mat)
             feats, coors = ff(feats, coors)
